@@ -20,7 +20,13 @@ except Exception as err:
         raise
 
 try:
-    from libifstate.xdp import XDP
+    from libifstate.bpf import libbpf, BPF
+
+    try:
+        from libifstate.xdp import XDP
+    except ModuleNotFoundError:
+        # ignore missing plugin
+        pass
 except ModuleNotFoundError:
     # ignore missing plugin
     pass
@@ -37,7 +43,7 @@ import json
 import errno
 import logging
 
-__version__ = "1.6.1"
+__version__ = "1.8.1"
 
 
 class IfState():
@@ -45,6 +51,7 @@ class IfState():
         logger.debug('IfState {}'.format(__version__))
         self.links = {}
         self.addresses = {}
+        self.bpf_progs = None
         self.neighbours = {}
         self.ignore = {}
         self.vrrp = {
@@ -59,6 +66,7 @@ class IfState():
         self.wireguard = {}
         self.xdp = {}
         self.features = {
+            'brport': True,
             'link': True,
             'sysctl': os.access('/proc/sys/net', os.R_OK),
             'ethtool': not ethtool_path is None,
@@ -109,6 +117,16 @@ class IfState():
                         self.sysctl.add(
                             iface, ifstates['options']['sysctl'][iface])
 
+        # load BPF programs
+        if 'bpf' in ifstates:
+            if not self.features['bpf']:
+                raise FeatureMissingError("bpf")
+
+            if self.bpf_progs is None:
+                self.bpf_progs = BPF()
+            for name, config in ifstates['bpf'].items():
+                self.bpf_progs.add(name, config)
+
         # add interfaces from config
         for ifstate in ifstates['interfaces']:
             name = ifstate['name']
@@ -116,7 +134,7 @@ class IfState():
                 raise LinkDuplicate()
             if 'link' in ifstate:
                 self.links[name] = Link(
-                    name, ifstate['link'], ifstate.get('ethtool'), ifstate.get('vrrp'))
+                    name, ifstate['link'], ifstate.get('ethtool'), ifstate.get('vrrp'), ifstate.get('brport'))
             else:
                 self.links[name] = None
 
@@ -274,6 +292,9 @@ class IfState():
                 logger.info("\nconfiguring {} interface sysctl".format(iface))
                 self.sysctl.apply(iface, do_apply)
 
+        if not self.bpf_progs is None:
+            self.bpf_progs.apply(do_apply)
+
         for stage in range(2):
             if stage == 0:
                 logger.info("\nconfiguring interface links")
@@ -305,8 +326,7 @@ class IfState():
                     else:
                         deps = link.depends()
                         if all(x in applied for x in deps):
-                            self.sysctl.apply(name, do_apply)
-                            excpts = link.apply(do_apply)
+                            excpts = link.apply(do_apply, self.sysctl)
                             if excpts.has_errno(errno.EEXIST):
                                 retry = True
                             applied.append(name)
@@ -380,7 +400,7 @@ class IfState():
                     logger.debug('skipped due to no xdp settings', extra={
                                  'iface': name})
                 else:
-                    xdp.apply(do_apply)
+                    xdp.apply(do_apply, self.bpf_progs)
 
         if any(not x is None for x in self.addresses.values()):
             logger.info("\nconfiguring interface ip addresses...")
@@ -402,7 +422,7 @@ class IfState():
                     addresses.apply(self.ipaddr_ignore, self.ignore.get(
                         'ipaddr_dynamic', True), do_apply)
         else:
-            logger.info("\nno interface ip addressing to be applied")
+            logger.debug("\nno interface ip addressing to be applied")
 
         if any(not x is None for x in self.neighbours.values()):
             logger.info("\nconfiguring interface neighbours...")
@@ -423,7 +443,7 @@ class IfState():
                 else:
                     neighbours.apply(do_apply)
         else:
-            logger.info("\nno interface neighbours to be applied")
+            logger.debug("\nno interface neighbours to be applied")
 
         if not self.tables is None:
             self.tables.apply(self.ignore.get('routes', []), do_apply)
@@ -521,6 +541,8 @@ class IfState():
                 if not mtu is None and not mtu in [1500, 65536]:
                     ifs_link['link']['mtu'] = mtu
 
+                brport.BRPort.show(showall, ipr_link['index'], ifs_link)
+
                 ifs_links.append(ifs_link)
 
         routing = {
