@@ -1,5 +1,6 @@
 from libifstate.util import logger, ipr, IfStateLogging
 from libifstate.exception import ExceptionCollector, LinkTypeUnknown, netlinkerror_classes
+from libifstate.brport import BRPort
 from abc import ABC, abstractmethod
 import os
 import subprocess
@@ -83,7 +84,7 @@ class Link(ABC):
         return super().__new__(GenericLink)
         #raise LinkTypeUnknown()
 
-    def __init__(self, name, link, ethtool, vrrp):
+    def __init__(self, name, link, ethtool, vrrp, brport):
         self.cap_create = True
         self.cap_ethtool = False
         self.settings = {
@@ -92,6 +93,10 @@ class Link(ABC):
         self.settings.update(link)
         self.ethtool = None
         self.vrrp = vrrp
+        if brport:
+            self.brport = BRPort(name, brport)
+        else:
+            self.brport = None
         self.attr_map = {
             'kind': ['IFLA_LINKINFO', 'IFLA_INFO_KIND'],
         }
@@ -243,7 +248,7 @@ class Link(ABC):
     def match_vrrp_state(self, vrrp_type, vrrp_name, vrrp_state):
         return self.match_vrrp_select(vrrp_type, vrrp_name) and (vrrp_state in self.vrrp['states'])
 
-    def apply(self, do_apply):
+    def apply(self, do_apply, sysctl):
         excpts = ExceptionCollector(self.settings['ifname'])
         osettings = copy.deepcopy(self.settings)
 
@@ -280,20 +285,20 @@ class Link(ABC):
                     excpts.add('set', err, state='down', ifname='{}!')
 
             if self.cap_create and self.get_if_attr('kind') != self.settings['kind']:
-                self.recreate(do_apply, excpts)
+                self.recreate(do_apply, sysctl, excpts)
             else:
                 excpts.set_quiet(self.cap_create)
-                self.update(do_apply, excpts)
-                if self.cap_create and excpts.get_all():
+                self.update(do_apply, sysctl, excpts)
+                if self.cap_create and list(excpts.get_all(lambda x: x['op'] != 'brport')):
                     excpts.reset()
-                    self.recreate(do_apply, excpts)
+                    self.recreate(do_apply, sysctl, excpts)
         else:
-            self.create(do_apply, excpts)
+            self.create(do_apply, sysctl, excpts)
 
         self.settings = osettings
         return excpts
 
-    def create(self, do_apply, excpts, oper="add"):
+    def create(self, do_apply, sysctl, excpts, oper="add"):
         logger.info(
             oper, extra={'iface': self.settings['ifname'], 'style': IfStateLogging.STYLE_CHG})
 
@@ -301,17 +306,34 @@ class Link(ABC):
             " ".join("{}={}".format(k, v) for k, v in self.settings.items())))
         if do_apply:
             try:
+                # set state later
                 state = self.settings.pop('state', None)
+
+                # prevent altname conflict
+                self.prevent_altname_conflict()
+
+                # add link
                 ipr.link('add', **(self.settings))
                 self.idx = next(iter(ipr.link_lookup(
                     ifname=self.settings['ifname'])), None)
-                if not state is None and not self.idx is None:
-                    try:
-                        ipr.link('set', index=self.idx, state=state)
-                    except Exception as err:
-                        if not isinstance(err, netlinkerror_classes):
-                            raise
-                        excpts.add('set', err, state=state)
+
+                if self.idx is not None:
+                    # set sysctl settings if required
+                    sysctl.apply(self.settings['ifname'], do_apply)
+
+                    # set brport settings if required
+                    if self.brport:
+                        if self.brport.has_changes(self.idx):
+                            self.brport.apply(do_apply, self.idx, excpts)
+
+                    # set state if required
+                    if state is not None and not excpts.has_op('brport'):
+                        try:
+                            ipr.link('set', index=self.idx, state=state)
+                        except Exception as err:
+                            if not isinstance(err, netlinkerror_classes):
+                                raise
+                            excpts.add('set', err, state=state)
             except Exception as err:
                 if not isinstance(err, netlinkerror_classes):
                     raise
@@ -322,7 +344,7 @@ class Link(ABC):
             self.set_ethtool_state(
                 self.settings['ifname'], self.ethtool.keys(), do_apply)
 
-    def recreate(self, do_apply, excpts):
+    def recreate(self, do_apply, sysctl, excpts):
         logger.debug('has wrong link kind %s, removing', self.settings['kind'], extra={
                      'iface': self.settings['ifname']})
         if do_apply:
@@ -333,9 +355,9 @@ class Link(ABC):
                     raise
                 excpts.add('del', err)
         self.idx = None
-        self.create(do_apply, "replace")
+        self.create(do_apply, ifstate, excpts, "replace")
 
-    def update(self, do_apply, excpts):
+    def update(self, do_apply, sysctl, excpts):
         logger.debug('checking link', extra={'iface': self.settings['ifname']})
 
         old_state = self.iface['state']
@@ -370,10 +392,14 @@ class Link(ABC):
                             if self.ethtool[setting][option] != ethtool[setting].get(option):
                                 has_ethtool_changes.add(setting)
 
+        has_brport_changes = False
+        if self.brport:
+            has_brport_changes = self.brport.has_changes(self.idx)
+
         if has_link_changes:
             logger.debug('needs to be configured', extra={
                          'iface': self.settings['ifname']})
-            if old_state:
+            if old_state != 'down':
                 logger.debug('shutting down', extra={
                              'iface': self.settings['ifname']})
                 if do_apply:
@@ -383,26 +409,47 @@ class Link(ABC):
                         if not isinstance(err, netlinkerror_classes):
                             raise
                         excpts.add('set', err, state='down')
+
                 if not 'state' in self.settings:
                     self.settings['state'] = 'up'
 
+            # set sysctl settings
+            sysctl.apply(self.get_if_attr(
+                'ifname'), do_apply, self.settings['ifname'])
+
             self.set_ethtool_state(self.get_if_attr(
                 'ifname'), has_ethtool_changes, do_apply)
 
-            if self.get_if_attr('ifname') == self.settings['ifname']:
-                logger.info('change', extra={
-                            'iface': self.settings['ifname'], 'style': IfStateLogging.STYLE_CHG})
-            else:
+            has_ifname_change = self.get_if_attr('ifname') != self.settings['ifname']
+            if has_ifname_change:
                 logger.info('change (was {})'.format(self.get_if_attr('ifname')), extra={
                             'iface': self.settings['ifname'], 'style': IfStateLogging.STYLE_CHG})
+            else:
+                logger.info('change', extra={
+                            'iface': self.settings['ifname'], 'style': IfStateLogging.STYLE_CHG})
+
+            logger.debug("ip link set: {}".format(
+                " ".join("{}={}".format(k, v) for k, v in self.settings.items())))
             if do_apply:
+                # temp. remove special settings
+                state = self.settings.pop('state', None)
+                peer = self.settings.pop('peer', None)
+
+                if has_ifname_change:
+                    self.prevent_altname_conflict()
+
                 try:
-                    state = self.settings.pop('state', None)
                     ipr.link('set', index=self.idx, **(self.settings))
                 except Exception as err:
                     if not isinstance(err, netlinkerror_classes):
                         raise
-                    excpts.add('set', err, state=state)
+                    excpts.add('set', err, **(self.settings))
+
+                # restore settings
+                if state is not None:
+                    self.settings['state'] = state
+                if peer is not None:
+                    self.settings['peer'] = peer
 
                 try:
                     if not state is None:
@@ -413,10 +460,35 @@ class Link(ABC):
                     if not isinstance(err, netlinkerror_classes):
                         raise
                     excpts.add('set', err, state=state)
+
+            if has_brport_changes:
+                self.brport.apply(do_apply, self.idx, excpts)
         else:
+            # set sysctl settings
+            sysctl.apply(self.get_if_attr(
+                'ifname'), do_apply)
+
             self.set_ethtool_state(self.get_if_attr(
                 'ifname'), has_ethtool_changes, do_apply)
 
+            if has_brport_changes:
+                if old_state:
+                    logger.debug('shutting down', extra={
+                                 'iface': self.settings['ifname']})
+                    if do_apply:
+                        try:
+                            ipr.link('set', index=self.idx, state='down')
+                        except Exception as err:
+                            if not isinstance(err, netlinkerror_classes):
+                                raise
+                            excpts.add('set', err, state='down')
+                    if not 'state' in self.settings:
+                        self.settings['state'] = 'up'
+
+                    has_state_changes = self.settings['state'] == 'up'
+
+                self.brport.apply(do_apply, self.idx, excpts)
+
             if has_state_changes:
                 try:
                     ipr.link('set', index=self.idx,
@@ -437,8 +509,41 @@ class Link(ABC):
         for attr in self.attr_idx:
             if attr in self.settings:
                 deps.append(self.settings[attr])
+
+        if self.brport:
+            deps.extend(self.brport.depends())
+
         return deps
 
+    def prevent_altname_conflict(self):
+        '''
+        When renaming interfaces a interface name conflict could
+        happen on the IFLA_ALT_IFNAME property (Linux 5.5+). Remove
+        the altname from the other interface in such case, it may be
+        the same interface if it is renamed to one of it's altnames.
+        '''
+
+        logger.debug('checking altname conflict', extra={'iface': self.settings['ifname']})
+
+        # get link candidate having the ifname as altname
+        try:
+            link = next(iter(ipr.link('get', altname=self.settings['ifname'])), None)
+        except Exception as err:
+            if not isinstance(err, netlinkerror_classes):
+                raise
+            return
+
+        # pyroute2 may return the interface having ifname rather than
+        # altname set - only remove altname if it is set
+        properties = link.get_attr('IFLA_PROP_LIST')
+        if properties is not None and self.settings['ifname'] in properties.get_attrs('IFLA_ALT_IFNAME'):
+            logger.debug('  found: %s (%d)', link.get_attr('IFLA_IFNAME'), link['index'], extra={'iface': self.settings['ifname']})
+            try:
+                ipr.link('property_del', index=link['index'], altname=self.settings['ifname'])
+            except Exception as err:
+                if not isinstance(err, netlinkerror_classes):
+                    raise
+
     @classmethod
     def name2nla(self, name):
         '''
@@ -455,5 +560,5 @@ class Link(ABC):
 
 
 class GenericLink(Link):
-    def __init__(self, name, link, ethtool, vrrp):
-        super().__init__(name, link, ethtool, vrrp)
+    def __init__(self, name, link, ethtool, vrrp, brport):
+        super().__init__(name, link, ethtool, vrrp, brport)
